Class('Symbie.Meta.Route.Token', {
    
    has : {
        token               : null,
        
        specificity         : 0
    },
    
    
    methods : {
        
        match : function (token) {
            throw "Abstract method 'match' was called on [" + this + "]"
        },
        
        
        asString : function () {
            throw "Abstract method 'asString' was called on [" + this + "]"
        }
    }
    
})


;
Class('Symbie.Meta.Route.Token.Parameter', {
    
    isa : 'Symbie.Meta.Route.Token',
    
    has : {
        name                : null,
        
        specificity         : 50,
        
        regex               : { required : true }
    },
    
    
    methods : {
        
        match : function (token) {
            return this.regex.exec(token)
        },
        
        
        asString : function (parameters) {
            var parameterValue = parameters && parameters[this.name]
            
            if (!parameterValue) throw "Parameter [" + this.token + "] wasn't supplied during token stringification"
            
            if (!this.regex.test(parameterValue)) throw "Supplied parameter [" + parameterValue + "] don't match the required regular expression in 'where'"
            
            return parameterValue
        }
    }
    
})


;
Class('Symbie.Meta.Route.Token.String', {
    
    isa : 'Symbie.Meta.Route.Token',
    
    has : {
        specificity         : 100
    },
    
    
    methods : {
        
        match : function (token) {
            return token == this.token
        },
        
        
        asString : function () {
            return this.token
        }
    }
    
})


;
Class('Symbie.Meta.Route.Token.Root', {
    
    isa : 'Symbie.Meta.Route.Token.String',
    
    has : {
        token               : ''
    }
    
})


;
Class('Symbie.Meta.Route.Token.WildCard', {
    
    isa : 'Symbie.Meta.Route.Token',
    
    has : {
        token               : '*',
        
        specificity         : 25
    },
    
    
    methods : {
        
        match : function (token) {
            return true
        },
        
        
        asString : function (parameters) {
            var splat       = parameters.splat
            
            if (splat instanceof Array) return splat.join('/')
            
            return ''
        }
    }
    
})


;
Class('Symbie.Meta.Route', {
    
    meta    : Joose.Meta.Class,
    
    isa     : Joose.Managed.Property,
        
    trait   : 'JooseX.CPS',
    
    use     : [ 
        'Symbie.Meta.Route.Token.String', 
        'Symbie.Meta.Route.Token.Root', 
        'Symbie.Meta.Route.Token.Parameter', 
        'Symbie.Meta.Route.Token.WildCard' 
    ],
    
    
    has : {
        action                  : { required : true },
        
        map                     : null,
        
        where                   : null,
        
        
        
        tokens                  : null,
        use                     : null,
        ready                   : false
    },
    
    
    after : {
        
        initialize : function () {
            var map     = this.map  = this.map != null ? this.map : (this.name == 'INDEX' ? '' : this.name) 
            
            this.tokens = this.stringToTokens(map)
        }
    },
    
    
    methods : {
        
        prepareApply : function (target) {
        },
        
        
        apply : function (target) {
        },
        
        
        unapply : function (from) {
        },
        
        
        stringToTokens : function (path) {
            var me          = this
            var tokens      = []
            var segments    = path.replace(/\/+/, '/').split('/')
            
            Joose.A.each(segments, function (token, index) {
                
                if (token == '.' && !index) return

                tokens.push(me.createToken(token, index, segments.length))
            })
            
            return tokens
        },
        
        
        isIndex : function () {
            var lastToken = this.tokens[ this.tokens.length - 1 ]
            
            return lastToken instanceof Symbie.Meta.Route.Token.String && lastToken.token == ''
        },
        
        
        // route can be local, so this method should return tokens for passed prefix, which forms a absolute path
        // prefix will also have trailing `/` symbol, so remove it to avoid extra token
        getTokensFor : function (prefix) {
            if (this.tokens[0] instanceof Symbie.Meta.Route.Token.Root) return [].concat(this.tokens)
            
            return this.stringToTokens(prefix.replace(/\/$/, '')).concat(this.tokens)
        },
        
        
        createToken : function (token, index, length) {
            if (token == '' && !index && length > 1)    return new Symbie.Meta.Route.Token.Root()
            
            if (/^\*/.test(token))                      return new Symbie.Meta.Route.Token.WildCard()
            
            if (/^:/.test(token)) {
                var name = /^:(.*)/.exec(token)[1]
                
                var regex = this.where && this.where[name] || /(.*)/
                
                return new Symbie.Meta.Route.Token.Parameter({ token : token, name : name, regex : regex })
            }
            
            return new Symbie.Meta.Route.Token.String({ token : token })
        },
        
        
        asString : function (parameters) {
            var res = []
            
            Joose.A.each(this.tokens, function (token, index) {
                var str = token.asString(parameters)
                
                if (str != null) res.push(str)
            })
            
            return res.join('/')
        },
        
        
        doExecute : function (context) {
            var args = [ context ]
            
            Joose.A.each(context.getParamsOrder(), function (paramName) {
                
                args.push(context.getParam(paramName))
            })
            
            return this.execute.apply(this, args)
        }
    },
    
    
    continued : {
        
        methods : {
            
            prepare : function () {
                var CONTINUE = this.getCONTINUE()
                
                var me      = this
                var deps    = this.use
                
                if (deps)
                    use(deps, function () {
                        
                        me.ready = true
                        
                        CONTINUE()
                    })
                else {
                    me.ready = true
                    
                    CONTINUE()
                }
            },
            
            
            execute : function (context) {
                if (!this.ready) this.prepare()   // async
                
                this.TRY(this.action, context.controller, arguments).now()
            }
        }
    }
})
;
Class('Symbie.Meta.Route.Match', {
    
    has : {
        tokens                  : { required : true },
        route                   : { required : true },
        controller              : { required : true },
        
        currentTokenIndex       : 0,
        
        parameters              : Joose.I.Object,
        paramsOrder             : Joose.I.Array
    },
    
    
    methods : {
        
        asString : function (params) {
            var str = this.route.asString(params)
            
            if (/^\//.test(str)) return str
            
            return this.controller.getFullPrefix() + str
        },
        
        
        consumeToken : function (token, index, length) {
            var currentToken    = this.tokens[ this.currentTokenIndex ]
            var parameters      = this.parameters
            var paramsOrder     = this.paramsOrder
            
            
            if (currentToken instanceof Symbie.Meta.Route.Token.WildCard) {
                if (!parameters.splat) {
                    parameters.splat    = []
                    
                    paramsOrder.push('splat')
                }
                
                parameters.splat.push(token)
                
                return true
            }
            
            this.currentTokenIndex++
            
            if (currentToken instanceof Symbie.Meta.Route.Token.String)
                return currentToken.match(token)
            
            if (currentToken instanceof Symbie.Meta.Route.Token.Parameter) {
                var match = currentToken.match(token)
                
                if (match) {
                    var paramName = currentToken.name
                    
                    
                    paramsOrder.push(paramName)
                    
                    if (match[1] != null && match[2] != null) 
                        parameters[paramName] = match
                    else
                        if (match[1] != null)
                            parameters[paramName] = match[1]
                        else
                            parameters[paramName] = token
                }
                
                return match
            }
            
            return false
        },
        
        
        compareBySpecificity : function (another, pathLength) {
            
            var thisTokens      = this.tokens
            var anotherTokens   = another.tokens
            
            var thisLength      = thisTokens.length
            var anotherLength   = anotherTokens.length
            
            // weight index routes with extra token from actual path a bit less 
            if (thisLength == pathLength && anotherLength == pathLength + 1 && another.route.isIndex())
                return 5
            
            if (anotherLength == pathLength && thisLength == pathLength + 1 && this.route.isIndex())
                return -5
            
            if (thisLength < anotherLength) return -1
            if (thisLength > anotherLength) return 1
            
            for (var i = 0; i < thisTokens.length; i++ ) {
                
                var specDiff = thisTokens[i].specificity - anotherTokens[i].specificity
                
                if (specDiff) return specDiff
            }
            
            return 0
        }
    }
    
})
;
Class('Symbie.Context', {
    
    trait : 'JooseX.CPS',
    
    has : {
        dispatchedFrom          : { required : true },
        dispatchParams          : { required : true },
        
        controller              : { required : true },
        match                   : { required : true },
        
        path                    : null,
        query                   : null,
        
        compChain               : Joose.I.Array,
        
        stash                   : Joose.I.Object
    },
    
    
    methods : {
        
        get : function (name) {
            return this.stash[ name ]
        },
        
        
        set : function (name, value) {
            this.stash[ name ] = value
        },
        
        
        app : function () {
            return this.getRoot()
        },
        
        
        getRoot : function () {
            return this.compChain[0]
        },
        
        
        initialize : function () {
            this.controller.getParents(this.compChain)
        },
        
        
        getParameters : function () {
            return this.match.parameters
        },
        
        
        getParams : function () {
            return this.getParameters()
        },
        
        
        getParamsOrder : function () {
            return this.match.paramsOrder
        },
        
        
        getParam : function (name) {
            return this.getParameters()[ name ]
        },
        
        
        getSplat : function () {
            return this.match.path
        },
        
        
        getRoute : function () {
            return this.match.route
        },
        
        
        eachComp : function (func, scope) {
            return Joose.A.each(this.compChain, func, scope || this)
        },
        
        
        eachCompR : function (func, scope) {
            var compChain = this.compChain
            
            for (var i = compChain.length - 1; i >= 0; i--)
                if (func.call(scope || this, compChain[i], i) === false) return false
        },
        
        
        runACTIVATE : function () {
            
            this.eachComp(function (controller) {
                
                controller.ACTIVATE(this)
            })
        },
        
        
        runFINALIZE : function () {
            
            this.eachCompR(function (controller) {
                
                controller.FINALIZE(this)
            })
        }
    },
    
    
    continued : {
        
        methods : {
            
            call : function (params) {
                var dispatchedFrom      = this.dispatchedFrom
                
                var match               = dispatchedFrom.findMatch(params)
                
                match.route.doExecute(this).now()
            },
            
            
            visit : function (path) {
//                this.redirected = true
//                
//                var me = this
//                
//                this.router.dispatch({
//                    routePath       : path,
//                    prevContext     : this
//                }).andThen(function () {
//                    
//                    this.CONTINUE(me)
//                })
            },
            
            
            redirect : function (path) {
//                this.redirected = true
//                
//                var me = this
//                
//                this.router.dispatch({
//                    routePath       : path,
//                    prevContext     : this
//                }).andThen(function () {
//                    
//                    this.CONTINUE(me)
//                })
            },
            
            
            run : function () {
                var me          = this
                var route       = this.getRoute()
                var controller  = this.controller
                
                this.runPRE().andThen(function (stop) {
                    
                    if (stop) {
                        this.CONTINUE(me)
                        
                        return
                    }
                    
                    this.runACTIVATE()      // sync
                    
                    controller.BEGIN(this)  // async
                    
                    route.doExecute(this)     // async
                    
                    controller.END(this).andThen(function () {
                        
                        me.runFINALIZE()    // sync
                        
                        this.CONTINUE(me)
                    })
                })
            },
            
            
            runPRE : function () {
                
                this.eachComp(function (controller) {
                    
                    controller.PRE(this).then(function (stop) {
                        
                        if (stop === false) 
                            this.RETURN(true)
                        else
                            this.CONTINUE()
                    })
                })
                
                this.NOW()
            }
        }
    }
    
});
Role('Symbie.Meta.Controller', {
    
    use : [ 'Symbie.Meta.Route' ],
    
    
    has : {
        prefix          : '',
        
        controllers     : Joose.I.Object,
        config          : Joose.I.Object,
        
        routes          : null,
        routesMC        : Joose.I.FutureClass('Symbie.Meta.Route')
    },
    
    
    methods : {
        
        hasRoute : function (name) {
            return this.stem.properties.routes.haveProperty(name)
        },
        
        
        hasOwnRoute : function (name) {
            return this.stem.properties.routes.haveOwnProperty(name)
        },
        
        
        getRoute : function (name) {
            return this.stem.properties.routes.getProperty(name)
        },
        
        
        getRoutes : function () {
            return this.stem.properties.routes
        }
    },
    
    
    after : {
        
        initInstance : function (instance, config) {
            var controllers = instance.controllers
            
            Joose.O.each(this.controllers, function (params, name) {
                
                controllers[ name ] = new (Joose.S.strToClass(name))(Joose.O.copy(params, { parent : instance }))
            })
        }
    },
    
    
    before : {
        
        // move the information from `plugins` and `controllers` to meta 
        prepareProps : function (props) {
            var does    = props.does    = [].concat(props.does || [])
            
            var config  = this.config   = props.config || {}
            delete props.config
            
            Joose.A.each(Joose.O.wantArray(props.plugins || []), function (pluginsDesc) {
                
                if (typeof pluginsDesc == 'string')
                    does.push(Joose.S.strToClass(pluginsDesc))
                else
                    Joose.O.each(pluginsDesc, function (params, name) {
                        does.push(Joose.S.strToClass(name))
                        
                        config[ name ] = params
                    })
            })
            
            delete props.plugins
            
            var controllers = this.controllers
            
            Joose.A.each(Joose.O.wantArray(props.controllers || []), function (value) {
                
                if (typeof value == 'string')
                    controllers[ value ] = {}
                else
                    Joose.O.each(value, function (params, name) {
                        controllers[ name ] = params
                    })
            })
            
            delete props.controllers
        },
        
        
        processStem : function () {
            var superMeta = this.superClass.meta
            
            this.routes = Joose.O.getMutableCopy(superMeta.routes || {})
        }
    },
    
    
    stem : {
        
        has     : {
            controllerInitialized       : false
        },
        
        // modifier can be applied several times
        after : {
            
            initialize : function () {
                if (!this.controllerInitialized) {
                    
                    this.controllerInitialized  = true
                    
                    // not just push, because this property will be actually taken from the prototype, so we shouldn't modify it
                    this.processOrder           = this.processOrder.concat('routes')
                    
                    var targetMeta              = this.targetMeta
                    
                    //will be Joose.Managed.PropetySet.Mutable
                    this.addProperty('routes', {
                        properties : targetMeta ? targetMeta.routes : {}
                    })
                }
            }
        }
    },
    
    
    builder : {
        
        methods : {
            
            routes : function (meta, info) {
                var routes = meta.stem.properties.routes
                
                Joose.O.each(info, function (props, name) {
                    if (typeof props == 'function') props = {
                        action : props
                    }
                    
                    props.meta = props.meta || meta.routesMC
                    
                    routes.addProperty(name, props)
                })
            },
            
            
            prefix  : function (meta, info) {
                meta.prefix = info
            }
        }
    }
})



JooseX.Namespace.Depended.meta.extend({
    
    methods : {
        
        symbieCollect : function (data, deps) {
            Joose.A.each(Joose.O.wantArray(data), function (desc) {
                
                if (typeof desc == 'string')
                    deps.push(desc)
                else
                    Joose.O.each(desc, function (params, name) {
                        deps.push(name)
                    })
            })
        }
        
    },
    
    
    after : {
        
        alsoDependsFrom : function (extend, summaredDeps) {
            
            if (extend.plugins)         this.symbieCollect(extend.plugins, summaredDeps)
            if (extend.controllers)     this.symbieCollect(extend.controllers, summaredDeps)
        }
    }
})
;
Class('Symbie.Controller', {
    
    traits      : [ 
        'JooseX.CPS',
        
        'Symbie.Meta.Controller'
    ],
    

    use         : [ 
        'Symbie.Context', 
        'Symbie.Meta.Route.Match' 
    ],
    
    
    does    : [
        'JooseX.Observable'
    ],
    
    
    has : {
        prefix                  : { is : 'rw' },
        fullPrefix              : { is : 'rw' },
        
        parent                  : { required : true },
        
        // will be filled by meta
        controllers             : Joose.I.Object,
        
        contextClass            : Joose.I.FutureClass('Symbie.Context')
    },
    
    
    methods : {
        
        getConfigFor : function (name) {
            return this.meta.config[ name ]
        },
        
        
        getParents : function (res) {
            res.unshift(this)
            
            if (this.parent) this.parent.getParents(res)
        },
        
        
        getRoot : function () {
            return this.parent && this.parent.getRoot() || this
        },
        
        
        // default value for prefix is being taken from meta
        // leading `./` is being stripped
        // missing trailing `/` is being added
        getPrefix : function () {
            var prefix      = this.prefix || this.meta.prefix
            
            return prefix.replace(/^\.\//, '').replace(/\/?$/, '/')
        },
        
        
        getFullPrefix : function () {
            var ownPrefix   = this.getPrefix()
            
            if (/^\//.test(ownPrefix)) return ownPrefix
            
            var prefix      = this.parent && this.parent.getFullPrefix() || '/'
            
            return prefix + ownPrefix
        },
        
        
        getOwnRoute : function (name) {
            return this.meta.getRoute(name)
        },
        
        
        getOwnRoutes : function () {
            return this.meta.getRoutes()
        },
        
        
        forEachController : function (func, scope) {
            return Joose.O.each(this.controllers, func, scope || this)
        },
        
        
        forEachRoute : function (func, scope) {
            var me      = this
            var self    = scope || this
            
            if (this.getOwnRoutes().each(function (route, name) {
                
                return func.call(self, route, name, me)
                
            }) === false) return false
            
            
            return this.forEachController(function (controller) {
                
                return controller.forEachRoute(func, scope)
            })
        },
        
        
        
        getMatchesPopulation : function (maxLength) {
            var matches = []
            
            this.forEachRoute(function (route, name, controller) {
                
                var tokens  = route.getTokensFor(controller.getFullPrefix())
                
                // add routes which has not much than `maxLength` tokens to population
                // exception is the index routes, which are allowed to have an extra token at the end
                if (tokens.length <= maxLength || tokens.length == maxLength + 1 && route.isIndex()) matches.push(new Symbie.Meta.Route.Match({
                    
                    controller      : controller,
                    tokens          : tokens,
                    route           : route    
                }))
                
            }, this)
            
            return matches
        },
        
        
        filterMatches : function (token, matches, index, length) {
            
            var filtered = []
            
            Joose.A.each(matches, function (match) {
                
                var survived = match.consumeToken(token, index, length)
                
                if (survived) filtered.push(match)
            })
            
            return filtered
        },
        
        
        findMatches : function (path) {
            var tokens = path.split('/')
            
            var matches = this.getMatchesPopulation(tokens.length)
            var length  = tokens.length
            
            Joose.A.each(tokens, function (token, index) {
                
                matches = this.filterMatches(token, matches, index, length)
            }, this)
            

            return matches
        },
        
        
        findMatch : function (path) {
            var matches     = this.findMatches(path)
            var length      = path.split('/').length
            
            if (!matches.length) throw "Can't find route for the path [" + path + "]"
            
            matches.sort(function (matchA, matchB) {
                return matchB.compareBySpecificity(matchA, length)
            })
                
            if (matches.length > 1 && matches[0].compareBySpecificity(matches[1], length) == 0) throw "Ambiguous route match"
                
            return matches[0]
        },
        
        
        parseQuery : function (queryStr) {
            var obj     = {}
            
            Joose.A.each(queryStr.split('&'), function (pair) {
                if (!pair) return
                
                var res = /([^=]*)(?:=(.*))?/.exec(pair)
                
                var name    = decodeURIComponent(res[1])
                var value   = decodeURIComponent(res[2]) || ''
                
                var already = obj[ name ]
                
                obj[ name ] = already != null ? [].concat(already, value) : value
            })
            
            return obj
        },
        
        
        ACTIVATE : function () {
        },
        
        
        FINALIZE : function () {
        }
    },
    
    
    continued : {
    
        methods : {

            PRE     : function () {
                this.CONTINUE()
            },
            
            
            BEGIN   : function () {
                this.CONTINUE()
            },
            
            
            END     : function () {
                this.CONTINUE()
            },
            
            
            dispatch : function (params) {
                if (typeof params == 'string') params = { path : params }
                
                //extracting trailing parameteres after '?'
                var res = /([^\?]*)(?:\?(.*))?/.exec(params.path)
                
                var path    = res[1]
                var query   = this.parseQuery(res[2] || '')
                
                try {
                    var match = this.findMatch(path)
                } catch (e) {
                    
                    if (this.fireEvent('dispatchException', e) !== false) 
                        throw e
                    else {
                        this.CONTINUE()
                        
                        return
                    }
                }
                
                var context = new this.contextClass({
                    dispatchParams      : params,
                    dispatchedFrom      : this,
                    
                    controller          : match.controller,
                    match               : match,
                    
                    query               : query,
                    path                : path
                })
                
                context.run().CATCH(function (e) {
                    
                    if (this.fireEvent('dispatchException', e) !== false) 
                        throw e
                    else
                        this.CONTINUE(context)
                    
                }, this).now()
            },
            
            
            setup : function () {
                
                this.forEachController(function (controller) {
                    
                    this.AND(function () {
                        controller.setup().now()
                    })
                })
                
                this.andTHEN(function () {
                    
                    this.fireEvent('ready', this)
                    
                    this.CONTINUE()
                })
            }
        }
    }
})
;
Class('Symbie.Application', {
    
    isa     : 'Symbie.Controller',
    

    has : {
        parent                  : null,
        prefix                  : '/'
    },
    
    
    methods : {
        
    },
    
    
    continued : {
        
        methods : {
            
            launch : function (path) {
                this.dispatch(path).now()
            },
            
            
            run : function (path) {
                this.setup()        // async
                
                this.launch(path || '/').now()
            }
        }
    }
    
});
Class('Symbie', {
    
    use     : [ 'Symbie.Application' ]
    
});
